from typing import TypeVar, Type, Any, Callable, get_origin, get_args
from functools import wraps
from types import UnionType, NoneType

T = TypeVar("T", bound=Type)


class FeatureToggler:
    """
    Class for disable methods.
    Usage:
        ```
        @FeatureToggler(enable=False)
        class PostgresManager:
        def __init__(self):
            self._postgres = create_engine('postgresql+psycopg2://user:password@127.0.0.1:5432/db')

        def test(self) -> int:
            with self._postgres.connect() as conn:
                conn.execute(text("SELECT 1;"))
            return 124  # Output: 0

        @FeatureToggler.ignore # Method ignored override
        def another_test(self) -> int:
            with self._postgres.connect() as conn:
                conn.execute(text('SELECT 2;'))
            return 777  # Output: 777
        ```
    Required: Annotation in functions!

    TODO Plans:
        1. add correct handling of cases when the method returns an instance of itself. Now will be None

    """

    __slots__ = ("_enable",)

    def __init__(self, enable: bool = True) -> None:
        self._enable: bool = enable

    def _default_value_by_type(self, return_type: Type) -> Any:
        """Return value by annotation."""
        returning_tuple_types: tuple = (return_type, get_origin(return_type))
        if int in returning_tuple_types:
            return 0
        if str in returning_tuple_types:
            return ""
        if list in returning_tuple_types:
            return []
        if dict in returning_tuple_types:
            return {}
        if bool in returning_tuple_types:
            return False
        if set in returning_tuple_types:
            return set()
        if frozenset in returning_tuple_types:
            return frozenset([])
        if bytes in returning_tuple_types:
            return bytes([])
        if UnionType in returning_tuple_types:
            return self._process_union_type(return_type)
        return None

    def _process_union_type(self, return_type: Type) -> Any:
        args: tuple[Any, ...] = get_args(return_type)
        if NoneType in args:
            return None
        for arg in args:
            default_value = self._default_value_by_type(arg)
            if default_value is not None:
                return default_value
        return None

    def disable_method(self, method: Callable) -> Callable:
        @wraps(method)
        def wrapper(*args, **kwargs) -> Any:
            return_type: Type | None = method.__annotations__.get("return", None)  # type: ignore
            return self._default_value_by_type(return_type)

        return wrapper

    def disable_property(self, prop: property) -> property:
        @property
        @wraps(prop.fget)
        def wrapper(*_) -> Any:
            return_type: Type | None = prop.fget.__annotations__.get("return", None)  # type: ignore
            return self._default_value_by_type(return_type)

        return wrapper

    def enable_method_by_dependency(self, enable: bool = True) -> Callable:
        def wrapper(func):
            if enable is False:
                return self.disable_method(func)
            return func

        return wrapper

    def __call__(self, cls: T) -> T:
        if not self._enable:
            for attr_name in dir(cls):
                attr = getattr(cls, attr_name)
                if attr_name.startswith("__") and attr_name.endswith("__"):
                    if not getattr(attr, "_force_override", False):
                        continue
                if callable(attr):
                    if getattr(attr, "_ignore", False):
                        continue
                    setattr(cls, attr_name, self.disable_method(attr))
                elif isinstance(attr, property):
                    if getattr(attr.fget, "_ignore", False):
                        continue
                    setattr(cls, attr_name, self.disable_property(attr))
        return cls

    @staticmethod
    def ignore(method: Callable) -> Any:
        """
        Wrapper for mark ignored methods
        """
        method._ignore = True
        return method

    @staticmethod
    def force_override(method: Callable) -> Any:
        """
        Wrapper for include dunder methods to override (default excludes)
        """
        method._force_override = True
        return method



from enum import Enum


class StrEnum(str, Enum):
    """
    Базовый класс для перечислений строковых значений.
    Наследуется от str и Enum, чтобы обеспечить строковое поведение.
    """

    def __str__(self):
        return str(self.value)



import contextlib
from base64 import b64encode

from gigachat.models import Embedding
from tqdm import tqdm
from gigachat import GigaChat
import typing as t
from itertools import chain
from time import sleep
from tqdm.contrib.concurrent import process_map


class GigaChatEntryPoint:
    __slots__ = ("_creds", "_model", "_DIM", "_ZEROS", "_ERROR_MESSAGE", "_warmed_up")

    def __init__(self, client_id: str, client_secret: str, obligatory_warmup=False):
        self._creds: str = b64encode(f"{client_id}:{client_secret}".encode("utf-8")).decode("utf-8")
        self._model = GigaChat(
            credentials=self._creds,
            base_url="https://gigachat.devices.sberbank.ru/api/v1",
            scope="GIGACHAT_API_CORP",
            model="GigaChat-Pro",
            verify_ssl_certs=False,
            profanity_check=False,
        )
        self._DIM: int = 1024
        self._ZEROS: t.List[int] = [0 for _ in range(self._DIM)]
        self._ERROR_MESSAGE: str = ""
        self._warmed_up: bool = False
        if obligatory_warmup:
            self.warmup()

    def __call__(self) -> GigaChat:
        return self._model

    def get_response(self, sentence: str) -> str:
        with contextlib.suppress(Exception):
            return self._model.chat(sentence).choices[0].message.content
        return self._ERROR_MESSAGE

    def get_response_by_payload(self, payload: t.List[t.Dict[str, str]]) -> str:
        """payload: [{"role": "system", "content": system}, {"role": "user", "content": replica}]"""
        with contextlib.suppress(Exception):
            return self._model.chat({"messages": payload}).choices[0].message.content
        return self._ERROR_MESSAGE

    def get_embedding(self, sentence: str) -> t.List[float]:
        with contextlib.suppress(Exception):
            return self._model.embeddings(sentence).data[0].embedding
        return self._ZEROS

    def get_embeddings(self, sentences: t.List[str], request_limit=50) -> t.List[t.List[float]]:
        embeddings: t.List[t.List[float]] | None = None
        counter: int = 0
        while embeddings is None and counter < request_limit:
            with contextlib.suppress(Exception):
                items: t.List[Embedding] = self._model.embeddings(sentences).data
                embeddings = [item.embedding for item in items]
                break
            sleep(0.1)
            counter += 1
        if embeddings is not None:
            return embeddings
        return [self._ZEROS for _ in sentences]

    def get_more_embeddings(
        self,
        sentences: t.List[str],
        batch_size: int = 100,
        hide_progress_bar: bool = False,
        parallel: bool = False,
    ) -> t.List[t.List[float]]:
        batches: t.List[t.List[str]] = self.make_batches(sentences, size=batch_size)
        if not parallel:
            emb_batches = [self.get_embeddings(batch) for batch in tqdm(batches, disable=hide_progress_bar)]
        else:
            # TODO threads instead of processes
            emb_batches = process_map(
                self.get_embeddings,
                batches,
                chunksize=1,
                max_workers=16,
                disable=hide_progress_bar,
            )
        return list(chain.from_iterable(emb_batches))

    def count_tokens(self, sentences: t.List[str]) -> t.List[int]:
        return [count.tokens for count in self._model.tokens_count(sentences)]

    @staticmethod
    def make_batches(items: list, size: int = 500) -> t.List[t.List[str]]:
        slices = [(i * size, (i + 1) * size) for i in range(len(items) // size + 1)]
        return [items[st:ed] for st, ed in slices]

    def warmup(self) -> None:
        assert self.get_response("Прогрев") != self._ERROR_MESSAGE, "Нет доступа к ллм!"
        assert self.get_embedding("Прогрев") != self._ZEROS, "Нет доступа к ллм!"
        self._warmed_up = True

    def is_warmed_up(self) -> bool:
        return self._warmed_up


class GigaPlusEntryPoint(GigaChatEntryPoint):
    def __init__(self, client_id: str, client_secret: str) -> None:
        super().__init__(client_id, client_secret)
        self._model = GigaChat(
            credentials=self._creds,
            base_url="https://gigachat.devices.sberbank.ru/api/v1",
            scope="GIGACHAT_API_CORP",
            model="GigaChat-Plus",
            verify_ssl_certs=False,
            profanity_check=False,
        )


class GigaMaxEntryPoint(GigaChatEntryPoint):
    def __init__(self, client_id: str, client_secret: str) -> None:
        super().__init__(client_id, client_secret)
        self._model = GigaChat(
            credentials=self._creds,
            base_url="https://gigachat-preview.devices.sberbank.ru/api/v1/",
            scope="GIGACHAT_API_CORP",
            model="GigaChat-Max-preview",
            verify_ssl_certs=False,
            profanity_check=False,
        )



import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Protocol

from pythonjsonlogger import jsonlogger
from sys import stdout

from ..dig_ass.db import make_session_id
from ..extentions import SingletonMeta


class SomeLogger(Protocol):
    def error(self, msg: object, *args, **kwargs) -> None:
        pass


class JsonFormatter(jsonlogger.JsonFormatter):
    """
    additional json formatting possible
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.json_ensure_ascii = False


class LoggerSingleton(metaclass=SingletonMeta):
    __slots__ = ("_logger",)

    def __init__(self, logger_config: dict) -> None:
        """
        :param logger_config: dict
        Example:
            {
                "name": "foo",
                "level_common": "DEBUG",
                "stdout_json": False,
                "format_stdout": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                "level_stdout_handler": "DEBUG",
                "file_dir": "./logs",
                "format_file": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                "level_file_handler": "DEBUG"
            }
        """
        self._logger = logging.getLogger(name=logger_config["name"])
        self._logger.setLevel(logger_config["level_common"])
        self._init_stdout_handler(logger_config)
        # configure logging to file
        if logger_config.get("write_to_file"):
            self._init_file_handler(logger_config)

    def _init_stdout_handler(self, logger_config: dict) -> None:
        stdout_handler = logging.StreamHandler(stdout)
        # configure logging to stdout
        # stdout_json == True -> set json Formatter
        if logger_config.get("stdout_json"):
            stdout_formatter = JsonFormatter(logger_config["format_stdout"])
        else:
            stdout_formatter = logging.Formatter(logger_config["format_stdout"])

        stdout_handler.setFormatter(stdout_formatter)
        stdout_handler.setLevel(logger_config["level_stdout_handler"])
        self._logger.addHandler(stdout_handler)

    def _init_file_handler(self, logger_config: dict) -> None:
        log_file_dir = Path(logger_config["file_dir"])
        if not log_file_dir.exists():
            self._logger.warning(f"Folder {log_file_dir.absolute()} is not exist. Creating...")
            log_file_dir.mkdir()
        log_name: str = logger_config["name"]
        file: str = f"{log_name}_{make_session_id()}.json"
        log_file_path: Path = log_file_dir / file
        max_bytes: int = 100 * 1024 * 1024  # 100 MB
        file_handler = RotatingFileHandler(log_file_path, mode="a", maxBytes=max_bytes)
        json_formatter = JsonFormatter(logger_config["format_file"])
        file_handler.setFormatter(json_formatter)
        file_handler.setLevel(logger_config["level_file_handler"])
        self._logger.addHandler(file_handler)

    def get(self):
        return self._logger

    @property
    def logger(self) -> logging.Logger:
        return self._logger


def log_llm_error(
    text: str | None = None,
    vector: list[float] | None = None,
    model: str = "gigachat",
    logger: SomeLogger = None,
) -> None:
    if not logger:
        raise AttributeError("Logger in function log_llm_error is required")
    if text is not None and not text:
        logger.error(f"No response from {model}!!!")
        return None
    if vector is not None and all(not item for item in vector):
        logger.error(f"No response from {model} encoder!!!")
        return None
    return None



import json, requests


class YandexGPTEntryPoint:
    def __init__(self, token, folder_id):
        iam_token = requests.post(
            url="https://iam.api.cloud.yandex.net/iam/v1/tokens",
            data=json.dumps({"yandexPassportOauthToken": token}),
        ).json()["iamToken"]
        self.folder_id = folder_id
        self.text_url = "https://llm.api.cloud.yandex.net/foundationModels/v1/completion"
        self.emb_url = "https://llm.api.cloud.yandex.net:443/foundationModels/v1/textEmbedding"
        self.doc_uri = f"emb://{self.folder_id}/text-search-doc/latest"
        self.query_uri = f"emb://{self.folder_id}/text-search-query/latest"
        self.headers = {
            "Authorization": f"Bearer {iam_token}",
            "Content-Type": "application/json",
        }
        self.DIM = 256
        self.ZEROS = [0 for _ in range(self.DIM)]
        self.ERROR_MESSAGE = ""
        self.warmup()

    def get_response(self, total_input: str) -> str:
        data = {
            "model_uri": f"gpt://{self.folder_id}/yandexgpt/latest",
            "messages": [
                {
                    "role": "user",
                    "text": total_input,
                }
            ],
        }
        try:
            response = requests.post(url=self.text_url, headers=self.headers, json=data).json()
            return response["result"]["alternatives"][0]["message"]["text"]
        except:
            return self.ERROR_MESSAGE

    def get_embeddings(self, total_input, input_is_long=False) -> list:
        data = {
            "modelUri": self.doc_uri if input_is_long else self.query_uri,
            "text": total_input,
        }
        try:
            response = requests.post(url=self.emb_url, headers=self.headers, json=data).json()
            return response["embedding"]
        except:
            return self.ZEROS

    def warmup(self) -> None:
        assert self.get_response("Прогрев") != self.ERROR_MESSAGE, "Нет доступа к ллм!"
        assert self.get_embeddings("Прогрев") != self.ZEROS, "Нет доступа к ллм!"
        assert self.get_embeddings("Прогрев", input_is_long=True) != self.ZEROS, "Нет доступа к ллм!"



from datetime import datetime


def make_session_id() -> str:
    return f"{datetime.now():%y%m%d%H%M%S}"


def make_name(outer_context: dict, dirty=True, short=False):
    user_id = outer_context["UserId"]
    session_id = outer_context["SessionId"]
    client_id = outer_context["ClientId"]
    if short:
        return f"{user_id}_{session_id}_{client_id}"
    long = f"user_{user_id}_session_{session_id}_client_{client_id}"
    if dirty:
        return f"{long}.json"
    return f"{long}_clean.json"
